import pyglet
from pyglet import font
from pyglet.window import key

from cocos.director import director
from cocos.layer import Layer
from cocos.scene import Scene

from sprites import RobotSprite, TokenSprite, EvilSprite, PowerSprite
from layers import StatusLayer
from misc import BACKGROUND


class Instructions(Layer):
    def __init__(self):
        super(Instructions, self).__init__()

        self.text = pyglet.text.Label("""This is your robot. You didn't have much time to work on the details, so it's kind of simple. The "Input" holds the robot program. The robot will take one instruction at a time from the input, and place it on the "Stack". If it doesn't know what to do with the instruction it will simply leave it there.

One of the instructions your robot understands is MOVE. This is a instruction that uses the last 2 instructions in the stack: they should be numbers counting the number of steps in the X and Y direction. Press any key to see how the robot interprets the program in its input, R to repeat.
        """,
        multiline=True,
        font_size=12,
        x=10,
        y=director.get_window_size()[1]/2-52,
        halign=font.Text.LEFT,
        valign=font.Text.TOP,
        batch=self.batch)

        self.text.width = director.get_window_size()[0]-20

        self.step = 0

    def on_enter(self):
        width, height = director.get_window_size()
        self.robot = RobotSprite("1 1 MOVE", x=width/2, y=height/2)
        self.robot.update_status_layer()
        self.add(self.robot)

    def on_key_release(self, key_pressed, mod):
        if self.step == 0:
            self.step = 1
            self.robot.next()
        elif key_pressed == key.R:
            self.robot.delete()
            self.on_enter()
            self.robot.next()
        else:
            director.replace(Scene(BACKGROUND, GettingTokens(), StatusLayer()))


class GettingTokens(Layer):
    def __init__(self):
        super(GettingTokens, self).__init__()

        self.text = pyglet.text.Label("""Your robot's program is limited to a certain number of instructions. Sometimes this may not be enough to complete a certain level. Lucky for you the robot can pick new instructions and add them to its program, as you can see on this example.
        
Press any key, R to repeat.
        """,
        multiline=True,
        font_size=12,
        x=10,
        y=director.get_window_size()[1]/2-52,
        halign=font.Text.LEFT,
        valign=font.Text.TOP,
        batch=self.batch)

        self.text.width = director.get_window_size()[0]-20

        self.step = 0

    def on_enter(self):
        width, height = director.get_window_size()
        self.robot = RobotSprite("1 1 MOVE MOVE", x=width/2, y=height/2)
        self.robot.update_status_layer()
        self.add(self.robot)

        token = TokenSprite('-1 MUL', x=384, y=324)
        self.add(token)
        self.robot.tokens.append(token)

    def on_key_release(self, key_pressed, mod):
        if self.step == 0:
            self.step = 1
            self.robot.next()
        elif key_pressed == key.R:
            self.robot.delete()
            self.on_enter()
            self.robot.next()
        else:
            director.replace(Scene(BACKGROUND, Arithmetic(), StatusLayer()))


class Arithmetic(Layer):
    def __init__(self):
        super(Arithmetic, self).__init__()

        self.text = pyglet.text.Label("""Your robot can do basic arithmetics using the instructions ADD, SUB, MUL and DIV. Each one of these instructions uses the last two values from the stack. Here the robot will calculate (3+4) x (5-1).
        
Press any key, R to repeat.
        """,
        multiline=True,
        font_size=12,
        x=10,
        y=director.get_window_size()[1]/2-52,
        halign=font.Text.LEFT,
        valign=font.Text.TOP,
        batch=self.batch)

        self.text.width = director.get_window_size()[0]-20

        self.step = 0

    def on_enter(self):
        width, height = director.get_window_size()
        self.robot = RobotSprite("3 4 ADD 5 1 SUB MUL", x=width/2, y=height/2)
        self.robot.update_status_layer()
        self.add(self.robot)

    def on_key_release(self, key_pressed, mod):
        if self.step == 0:
            self.step = 1
            self.robot.next()
        elif key_pressed == key.R:
            self.robot.delete()
            self.on_enter()
            self.robot.next()
        else:
            director.replace(Scene(BACKGROUND, Variables(), StatusLayer()))


class Variables(Layer):
    def __init__(self):
        super(Variables, self).__init__()

        self.text = pyglet.text.Label("""You can also assign values to variables, using the DEF instruction. Defined variables can later be accessed by name instead of value. Note that there are two special variables called "x" and "y" that hold the robot's position.

In this program we associate "a" with an expression, delimited by braces. Note that you need to put a slash before "a", to identify the variable.
        
Press any key, R to repeat.
        """,
        multiline=True,
        font_size=12,
        x=10,
        y=director.get_window_size()[1]/2-52,
        halign=font.Text.LEFT,
        valign=font.Text.TOP,
        batch=self.batch)

        self.text.width = director.get_window_size()[0]-20

        self.step = 0

    def on_enter(self):
        width, height = director.get_window_size()
        self.robot = RobotSprite("/a { 640 2 DIV } DEF a x y", x=width/2, y=height/2)
        self.robot.update_status_layer()
        self.add(self.robot)

    def on_key_release(self, key_pressed, mod):
        if self.step == 0:
            self.step = 1
            self.robot.next()
        elif key_pressed == key.R:
            self.robot.delete()
            self.on_enter()
            self.robot.next()
        else:
            director.replace(Scene(BACKGROUND, Mission()))


class Mission(Layer):
    def __init__(self):
        super(Mission, self).__init__()

        self.text = pyglet.text.Label("""Well, that should get you started. I'm afraid we're running out of time, so you'll have to find out what other instructions do by experimentating with them.
        
Your mission is to collect 5 power generators, like the one on the left, one per level. Look out for evil robots like the one on the right, they are far more powerful than your robot -- they are made of an alien material called "metal".

Good luck! Now press any key to close the manual.
        """,
        multiline=True,
        font_size=12,
        x=10,
        y=director.get_window_size()[1]/2-52,
        halign=font.Text.LEFT,
        valign=font.Text.TOP,
        batch=self.batch)

        self.text.width = director.get_window_size()[0]-20

        self.step = 0

    def on_enter(self):
        width, height = director.get_window_size()

        evil = EvilSprite(x=width/2+96, y=height/2)
        self.add(evil)

        power = PowerSprite(x=width/2-96, y=height/2)
        self.add(power)

    def on_key_release(self, key_pressed, mod):
        director.pop()
