import sys
import re
from collections import deque
import md5
import math

import pyglet

from cocos.actions import *
from cocos.director import director


class EvilSprite(ActionSprite):
    def __init__(self, *args, **kwargs):
        image = pyglet.resource.image('evil.png')
        image.anchor_x = image.width / 2
        image.anchor_y = image.height / 2
        super(EvilSprite, self).__init__(image, *args, **kwargs)


class PowerSprite(ActionSprite):
    def __init__(self, *args, **kwargs):
        image = pyglet.resource.image('power.png')
        image.anchor_x = image.width / 2
        image.anchor_y = image.height / 2
        super(PowerSprite, self).__init__(image, *args, **kwargs)


class TokenSprite(ActionSprite):
    def __init__(self, text, *args, **kwargs):
        self.text = text

        hash = md5.md5(text).hexdigest()
        try:
            image = pyglet.resource.image('token_%s.png' % hash)
        except:
            image = self.create_image()
        image.anchor_x = image.width / 2
        image.anchor_y = image.height / 2
        super(TokenSprite, self).__init__(image, *args, **kwargs)

    def create_image(self):
        import Image, ImageDraw, ImageFont

        im = Image.new('RGB', (1, 1), (50, 50, 50))
        draw = ImageDraw.Draw(im)
        BORDER = 8
        w, h = draw.textsize(self.text)
        im = im.resize((w+2*BORDER, h+2*BORDER))
        draw = ImageDraw.Draw(im)
        ft = ImageFont.truetype('data/VeraMono.ttf', 10)
        draw.text((BORDER, BORDER), self.text, fill=(255, 255, 255), font=ft)

        base = Image.new('RGB', (w+2*BORDER+2, h+2*BORDER+2), (200, 200, 200))
        base.paste(im, (1, 1))

        hash = md5.md5(self.text).hexdigest()
        base.save('data/token_%s.png' % hash)
        pyglet.resource.reindex()
        return pyglet.resource.image('token_%s.png' % hash)


PASS = Delay(1)

class RobotSprite(ActionSprite):

    INSTRUCTIONS = ['MOVE', 'DEF', 'ADD', 'SUB', 'MUL', 'DIV', 'REPEAT',
            'DUP', 'POP', 'EXCH', 'ROLL', 'IF', 'IFELSE', 'EQ', 'NE',
            'GT', 'GE', 'LT', 'LE', 'OR', 'NOT']

    image = pyglet.resource.image('robot.png')
    image.anchor_x = image.width / 2
    image.anchor_y = image.height / 2

    def __init__(self, program, max_input=sys.maxint, *args, **kwargs):
        super(RobotSprite, self).__init__(self.image, *args, **kwargs)

        self.max_input = max_input
        if isinstance(program, basestring):
            program = re.split('[\s\n\r]+', program)
            program = program[:max_input]
        self.input = deque(program)

        self.stack = deque()
        self.dictionary = {}

        self.tokens = []
        self.enemies = []
        self.obstacles = []
        self.generators = []

        self.won = 0

    def step(self, dt):
        for i, token in enumerate(self.tokens):
            if self.collide(token):
                tokens = re.split('[\s\n\r]+', token.text)
                self.input.extendleft(reversed(tokens))
                token.delete()
                del self.tokens[i]
                self.update_status_layer()

        for generator in self.generators:
            if self.collide(generator):
                self.won = 1
                generator.delete()
                self.generators = []

                # Stop all actions 
                self.stop()
                for enemy in self.enemies:
                    enemy.stop()
                    enemy.do(Jump(10, 0, 3, 0.5))

        if self.won == 0:
            for enemy in self.enemies:
                if self.collide(enemy):
                    self.won = -1
                    self.input.clear()
                    self.stop()
                    self.do(Jump(10, 0, 3, 0.5))

        super(RobotSprite, self).step(dt)

    def collide(self, other):
        x0, y0 = self.x - self.image.width/2, self.y - self.image.height/2
        x1, y1 = self.x + self.image.width/2, self.y + self.image.height/2
        xp0, yp0 = other.x - other.image.width/2, other.y - other.image.height/2
        xp1, yp1 = other.x + other.image.width/2, other.y + other.image.height/2

        return ((x1-xp0) * (x0-xp1) < 0) and ((y1-yp0) * (y0-yp1) < 0)

    def update_status_layer(self):
        status_layer = director.scene.layers[-1][1]
        try:
            status_layer.input.text = 'Input: [ %s ]' % ' '.join(self.input)
            status_layer.stack.text = 'Stack: [ %s ]' % ' '.join(self.stack)
        except:
            pass

    def next(self):
        if not self.input: return

        self.dictionary.update(X=[str(self.x)], Y=[str(self.y)])
        token = self.input.popleft()
        if token == '{':
            lt, rt = 0, 0
            while 1:
                self.stack.append(token)
                if token == '{': lt += 1
                elif token == '}': rt += 1
                if lt == rt: break
                token = self.input.popleft()
            action = PASS
        elif token in self.INSTRUCTIONS:
            try:
                action = getattr(self, token)() or PASS
            except:
                self.stack.clear()
                action = Jump(10, 0, 3, 0.5)
        elif token in self.dictionary:
            self.input.extendleft(reversed(self.dictionary[token]))
            action = PASS
        else:
            self.stack.append(token)
            action = PASS

        self.do(Sequence(action, CallFunc(self.next)))
        self.update_status_layer()

    def MOVE(self):
        y = int(self.stack[-1])
        x = int(self.stack[-2])
        duration = math.sqrt(x**2+y**2)
        return Move((x*64, y*64), duration)

    def DEF(self):
        d = deque()
        lt, rt = 0, 0
        while 1:
            token = self.stack.pop()
            if token == '{': lt += 1
            elif token == '}': rt += 1
            elif token.startswith('/') and lt == rt: break
            d.appendleft(token)
        name = token[1:]
        if d[0] == '{':
            d.popleft()  # {
            d.pop()      # }
        self.dictionary[name] = d

    def IF(self):
        true = deque()
        lt, rt = 0, 0
        while 1:
            token = self.stack.pop()
            if token == '{': lt += 1
            elif token == '}': rt += 1
            true.append(token)
            if lt == rt: break
        if true[-1] == '{':
            true.popleft()
            true.pop()
        condition = self.stack.pop()
        if condition == 'TRUE':
            self.input.extendleft(true)
        elif condition != 'FALSE':
            raise Exception('Token is neither TRUE nor FALSE')

    def IFELSE(self):
        false = deque()
        lt, rt = 0, 0
        while 1:
            token = self.stack.pop()
            if token == '{': lt += 1
            elif token == '}': rt += 1
            false.append(token)
            if lt == rt: break
        if false[-1] == '{':
            false.popleft()
            false.pop()

        true = deque()
        lt, rt = 0, 0
        while 1:
            token = self.stack.pop()
            if token == '{': lt += 1
            elif token == '}': rt += 1
            true.append(token)
            if lt == rt: break
        if true[-1] == '{':
            true.popleft()
            true.pop()

        condition = self.stack.pop()
        if condition == 'TRUE':
            self.input.extendleft(true)
        elif condition == 'FALSE':
            self.input.extendleft(false)
        else:
            raise Exception('Token is neither TRUE nor FALSE')

    def REPEAT(self):
        d = deque()
        lt, rt = 0, 0
        while 1:
            token = self.stack.pop()
            if token == '{': lt += 1
            elif token == '}': rt += 1
            d.append(token)
            if lt == rt: break
        count = int(self.stack.pop())
        if d[-1] == '{':
            d.popleft()
            d.pop()
        for _ in range(count):
            self.input.extendleft(d)

    def ADD(self):
        y = int(self.stack.pop())
        x = int(self.stack.pop())
        self.stack.append(str(x+y))

    def SUB(self):
        y = int(self.stack.pop())
        x = int(self.stack.pop())
        self.stack.append(str(x-y))

    def MUL(self):
        y = int(self.stack.pop())
        x = int(self.stack.pop())
        self.stack.append(str(x*y))

    def DIV(self):
        y = int(self.stack.pop())
        x = int(self.stack.pop())
        self.stack.append(str(x/y))

    def DUP(self):
        token = self.stack.pop()
        self.stack.append(token)
        self.stack.append(token)

    def POP(self):
        self.stack.pop()

    def EXCH(self):
        y = self.stack.pop()
        x = self.stack.pop()
        self.stack.append(y)
        self.stack.append(x)

    def ROLL(self):
        self.stack.rotate(1)

    def EQ(self):
        y = int(self.stack.pop())
        x = int(self.stack.pop())
        result = str(x==y).upper()
        self.stack.append(result)

    def NE(self):
        y = int(self.stack.pop())
        x = int(self.stack.pop())
        result = str(x!=y).upper()
        self.stack.append(result)

    def GT(self):
        y = int(self.stack.pop())
        x = int(self.stack.pop())
        result = str(x>y).upper()
        self.stack.append(result)

    def GE(self):
        y = int(self.stack.pop())
        x = int(self.stack.pop())
        result = str(x>=y).upper()
        self.stack.append(result)

    def LT(self):
        y = int(self.stack.pop())
        x = int(self.stack.pop())
        result = str(x<y).upper()
        self.stack.append(result)

    def LE(self):
        y = int(self.stack.pop())
        x = int(self.stack.pop())
        result = str(x<=y).upper()
        self.stack.append(result)

    def OR(self):
        y = self.stack.pop()
        x = self.stack.pop()
        result = str(x == 'TRUE' or y == 'TRUE').upper()
        self.stack.append(result)

    def NOT(self):
        x = self.stack.pop()
        if x == 'TRUE': result = 'FALSE'
        elif x == 'FALSE': result = 'TRUE'
        else: raise Exception('Token is neither TRUE nor FALSE')
        self.stack.append(result)
